# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import os
from collections import defaultdict
from ast_canopy import (
    parse_declarations_from_source,
)

from numbast import (
    MemoryShimWriter,
    bind_cxx_structs,
    bind_cxx_functions,
)

from numba import types, config, cuda
from numba.core.datamodel.models import PrimitiveModel, StructModel

if cuda.get_current_device().compute_capability < (8, 0):
    import warnings

    warnings.warn(
        "bf16 bindings are only supported on compute capability 8.0 and later, "
        "most bf16 features may not be available.",
    )

CUDA_INCLUDE_PATH = config.CUDA_INCLUDE_PATH
COMPUTE_CAPABILITY = cuda.get_current_device().compute_capability

cuda_bf16 = os.path.join(CUDA_INCLUDE_PATH, "cuda_bf16.h")
cuda_bf16_hpp = os.path.join(CUDA_INCLUDE_PATH, "cuda_bf16.hpp")


structs, functions, _, _, typedefs, _ = parse_declarations_from_source(
    cuda_bf16,
    [cuda_bf16, cuda_bf16_hpp],
    f"sm_{COMPUTE_CAPABILITY[0]}{COMPUTE_CAPABILITY[1]}",
    cudatoolkit_include_dir=CUDA_INCLUDE_PATH,
)

TYPE_SPECIALIZATION = {
    "__nv_bfloat16_raw": types.Number,
    "__nv_bfloat16": types.Number,
    "__nv_bfloat162_raw": types.Type,
    "__nv_bfloat162": types.Type,
    "nv_bfloat16": types.Number,
    "nv_bfloat162": types.Type,
}
DATA_MODEL_SPECIALIZATION = {
    "__nv_bfloat16_raw": PrimitiveModel,
    "__nv_bfloat16": PrimitiveModel,
    "nv_bfloat16": PrimitiveModel,
    "__nv_bfloat162_raw": StructModel,
    "__nv_bfloat162": StructModel,
    "nv_bfloat162": StructModel,
}

functions_to_ignore = {
    "atomicAdd"  # TODO: build bindings to numba.cuda.atomic, not as a standalone function
}

numba_struct_types = []
numba_functions = []

aliases = defaultdict(list)
for typedef in typedefs:
    aliases[typedef.underlying_name].append(typedef.name)

# WAR for NVBug 4549708: ABI mismatching between the calling convention of the
# bf16 definition against the linked library, causing a unspecified launch failure.
pretext = f"""#define __FORCE_INCLUDE_CUDA_FP16_HPP_FROM_FP16_H__
#define __FORCE_INCLUDE_CUDA_BF16_HPP_FROM_BF16_H__
#include "{cuda_bf16}"
"""

shim_writer = MemoryShimWriter(pretext)

numba_struct_types += bind_cxx_structs(
    shim_writer, structs, TYPE_SPECIALIZATION, DATA_MODEL_SPECIALIZATION, aliases
)

numba_functions += bind_cxx_functions(
    shim_writer, functions, exclude=functions_to_ignore
)

# Export
globals().update({s.__name__: s for s in numba_struct_types})
globals().update({f.__name__: f for f in numba_functions})
for underlying_name, names in aliases.items():
    for name in names:
        if name not in globals() and underlying_name in globals():
            globals()[name] = globals()[underlying_name]
globals().update({"get_shims": shim_writer.links})

__all__ = list(  # noqa: F822
    set(s.__name__ for s in numba_struct_types)
    | set(f.__name__ for f in numba_functions)
    | set(typedef.name for typedef in typedefs)
) + ["get_shims"]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import operator
import re

from numba import cuda, types
from numba.extending import models, register_model, typeof_impl
from numba.cuda.cudadecl import register_global
from numba.cuda.cudaimpl import lower
from numba.core.typing.templates import AbstractTemplate, signature
import numpy as np

from llvmlite import ir

from numbast.types import CTYPE_MAPS

from ast_canopy.decl import Struct

# C-compatible type mapping, see:
# https://numpy.org/devdocs/reference/arrays.scalars.html#integer-types
ctystr_to_npystr = {
    "char": f"int{np.dtype(np.byte).itemsize * 8}",
    "short": f"int{np.dtype(np.short).itemsize * 8}",
    "int": f"int{np.dtype(np.intc).itemsize * 8}",
    "long": f"int{np.dtype(np.int_).itemsize * 8}",
    "long long": f"int{np.dtype(np.longlong).itemsize * 8}",
    "unsigned char": f"uint{np.dtype(np.ubyte).itemsize * 8}",
    "unsigned short": f"uint{np.dtype(np.ushort).itemsize * 8}",
    "unsigned int": f"uint{np.dtype(np.uintc).itemsize * 8}",
    "unsigned long": f"uint{np.dtype(np.uint).itemsize * 8}",
    "unsigned long long": f"uint{np.dtype(np.ulonglong).itemsize * 8}",
    "float": f"float{np.dtype(np.single).itemsize * 8}",
    "double": f"float{np.dtype(np.double).itemsize * 8}",
}

vector_type_base_to_npystr = {
    "char": f"int{np.dtype(np.byte).itemsize * 8}",
    "short": f"int{np.dtype(np.short).itemsize * 8}",
    "int": f"int{np.dtype(np.intc).itemsize * 8}",
    "long": f"int{np.dtype(np.int_).itemsize * 8}",
    "longlong": f"int{np.dtype(np.longlong).itemsize * 8}",
    "uchar": f"uint{np.dtype(np.ubyte).itemsize * 8}",
    "ushort": f"uint{np.dtype(np.ushort).itemsize * 8}",
    "uint": f"uint{np.dtype(np.uintc).itemsize * 8}",
    "ulong": f"uint{np.dtype(np.uint).itemsize * 8}",
    "ulonglong": f"uint{np.dtype(np.ulonglong).itemsize * 8}",
    "float": f"float{np.dtype(np.single).itemsize * 8}",
    "double": f"float{np.dtype(np.double).itemsize * 8}",
}


def to_numpy_dtype(ctystr: str):
    """Given a C type string, return the corresponding NumPy dtype.
    If the type is an array type, return a tuple of the base dtype and array size.
    Otherwise, return the base dtype and None.
    """
    vector_type_pattern = r"(.*)([1-4])$"
    vector_type_match = re.match(vector_type_pattern, ctystr)

    if vector_type_match:
        base_type, size = vector_type_match.groups()
        dtype = np.dtype(vector_type_base_to_npystr[base_type])
        return (dtype, int(size))
    elif "[" in ctystr:
        base_type = ctystr.split("[")[0]
        array_size = int(ctystr.split("[")[1].split("]")[0])
        dtype = np.dtype(ctystr_to_npystr[base_type])
        return (dtype, array_size)
    else:
        dtype = np.dtype(ctystr_to_npystr[ctystr])
        return (dtype,)


def make_curand_states(curand_state_decl: Struct):
    """Create a boxed curand states object to represent an array of curand state.
    Should invoke after creating the numba types for each curand state.
    """

    curand_state_name = curand_state_decl.name

    state_ty = CTYPE_MAPS[curand_state_name]

    curand_states_name = curand_state_name.replace("State", "States")

    # cuRAND state type as a NumPy dtype - this mirrors the state defined in
    # curand_kernel.h. Can be used to inspect the state through the device array
    # held by CurandStates.

    state_fields = [
        (f.name, *to_numpy_dtype(f.type_.name)) for f in curand_state_decl.fields
    ]

    curandStateDtype = np.dtype(state_fields, align=True)

    # Hold an array of cuRAND states - somewhat analogous to a curandState* in
    # C/C++.

    class CurandStates:
        def __init__(self, n):
            self._array = cuda.device_array(n, dtype=curandStateDtype)

        @property
        def data(self):
            return self._array.__cuda_array_interface__["data"][0]

    CurandStates.__name__ = curand_states_name

    class CurandStatePointer(types.Type):
        def __init__(self, name):
            self.dtype = state_ty
            super().__init__(name=name)

    curand_state_pointer = CurandStatePointer(curand_states_name + "*")

    @typeof_impl.register(CurandStates)
    def typeof_curand_states(val, c):
        return curand_state_pointer

    @register_global(operator.getitem)
    class CurandStatesPointerGetItem(AbstractTemplate):
        def generic(self, args, kws):
            assert not kws
            [ptr, idx] = args
            if ptr == curand_state_pointer:
                return signature(
                    types.CPointer(state_ty), curand_state_pointer, types.int64
                )

    register_model(CurandStatePointer)(models.PointerModel)

    @lower(operator.getitem, curand_state_pointer, types.int64)
    def lower_curand_states_getitem(context, builder, sig, args):
        [ptr, idx] = args

        # Working out my own GEP
        ptrint = builder.ptrtoint(ptr, ir.IntType(64))
        itemsize = curandStateDtype.itemsize
        offset = builder.mul(idx, context.get_constant(types.int64, itemsize))
        ptrint = builder.add(ptrint, offset)
        ptr = builder.inttoptr(ptrint, ptr.type)
        return ptr

    # Argument handling. When a CurandStatePointer is passed into a kernel, we
    # really only need to pass the pointer to the data, not the whole underlying
    # array structure. Our handler here transforms these arguments into a uint64
    # holding the pointer.

    class CurandStateArgHandler:
        def prepare_args(self, ty, val, **kwargs):
            if isinstance(val, CurandStates):
                assert ty == curand_state_pointer
                return types.uint64, val.data
            else:
                return ty, val

    curand_state_arg_handler = CurandStateArgHandler()

    return CurandStates, curand_state_arg_handler



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import warnings
import numpy as np
import logging
import sys

import numba
from numba.np import numpy_support  # noqa: E402
import numba.cuda.dispatcher
from numba.cuda.dispatcher import _LaunchConfiguration

from numbast_extensions.bf16 import nv_bfloat16

_logger = None


# Create a logger that reports messages based on the value of Numba's config
# variable NUMBA_CUDA_LOG_LEVEL, so we can trace patching when tracing other
# CUDA operations.
def get_logger():
    global _logger

    if _logger:
        return _logger

    logger = logging.getLogger(__name__)

    # Create a default configuration if none exists already
    if not logger.hasHandlers():
        lvl = str(numba.config.CUDA_LOG_LEVEL).upper()
        lvl = getattr(logging, lvl, None)

        if not isinstance(lvl, int):
            # Default to critical level
            lvl = logging.CRITICAL
        logger.setLevel(lvl)

        # Did user specify a level?
        if numba.config.CUDA_LOG_LEVEL:
            # Create a simple handler that prints to stderr
            handler = logging.StreamHandler(sys.stderr)
            fmt = (
                "== CUDA (Numbast_extensions) [%(relativeCreated)d] "
                "%(levelname)5s -- %(message)s"
            )
            handler.setFormatter(logging.Formatter(fmt=fmt))
            logger.addHandler(handler)
        else:
            # Otherwise, put a null handler
            logger.addHandler(logging.NullHandler())

    _logger = logger
    return logger


try:
    import torch

    _TORCH_IMPORTED = True
except ImportError:
    _TORCH_IMPORTED = False

try:
    from ml_dtypes import bfloat16  # noqa: F401 E402

    _ML_DTYPES_IMPORTED = True
except ImportError:
    if _TORCH_IMPORTED:
        warnings.warn(
            "Pytorch installation detected. To use numbast-extensions' bfloat16"
            "bindings, please also install ml_dtypes via `python -m pip install ml_dtypes`"
        )

    _ML_DTYPES_IMPORTED = False

_WRAP_BF16_TENSOR = _TORCH_IMPORTED and _ML_DTYPES_IMPORTED


def patch_numba():
    """
    This function patches Numba to support bfloat16 data type from PyTorch.
    It creates a proxy object that implements the CUDA array interface (CAI) that provides
    correct interpretation to bfloat16 tensors. Then it patches the _LaunchConfiguration
    object in Numba to enable Numba to directly consume bfloat16 tensors.

    In order for this patching to work, PyTorch and ml_dtypes must be installed.

    Returns:
        None
    """
    if not _WRAP_BF16_TENSOR:
        return

    logger = get_logger()

    logger.debug(
        "Patching Numba's _LaunchConfiguration for torch bfloat16-tensor "
        "interoperability."
    )

    # Register the NumPy dtype for bfloat16 with the Numba bfloat16 type
    numpy_support.FROM_DTYPE[np.dtype("bfloat16")] = nv_bfloat16._nbtype

    class ProxyTorch(torch.Tensor):
        """
        Proxy object for PyTorch tensors that implements the CUDA array interface (CAI)
        to enable Numba to directly consume bfloat16 tensors.
        """

        def __init__(self, tensor):
            """
            Initialize the proxy object with the given PyTorch tensor.

            Args:
                tensor (torch.Tensor): The PyTorch tensor to wrap.
            """
            self._tensor = tensor

        def __getattr__(self, attr):
            """
            Delegate attribute access to the wrapped PyTorch tensor.

            Args:
                attr (str): The attribute name.

            Returns:
                The attribute value.
            """
            if attr == "__cuda_array_interface__":
                return self.__cuda_array_interface__

            return super(ProxyTorch, self).__getattr__(attr)

        @property
        def __cuda_array_interface__(self):
            """
            Implement the CUDA array interface for the proxy object.

            Returns:
                dict: A dictionary representing the CUDA array interface.
            """
            typestr = "bfloat16"

            if self._tensor.is_contiguous():
                # __cuda_array_interface__ v2 requires the strides to be omitted
                # (either not set or set to None) for C-contiguous arrays.
                strides = None
            else:
                strides = tuple(
                    s * torch.bfloat16.itemsize for s in self._tensor.stride()
                )
            shape = tuple(self.shape)
            data_ptr = self._tensor.data_ptr() if self._tensor.numel() > 0 else 0
            data = (data_ptr, False)  # read-only is false
            return dict(
                typestr=typestr, shape=shape, strides=strides, data=data, version=2
            )

    class _BF16TorchWrappedLaunchConfiguration(_LaunchConfiguration):
        """
        Custom launch configuration for Numba CUDA kernels that wraps bfloat16 PyTorch
        tensors with the proxy object before launching the kernel.
        """

        def __call__(self, *args):
            """
            Wrap PyTorch tensors with the proxy object before launching the kernel.

            Args:
                *args: The arguments to the kernel.

            Returns:
                The result of the kernel launch.
            """
            torch_filtered = [
                ProxyTorch(arg)
                if (isinstance(arg, torch.Tensor) and arg.dtype == torch.bfloat16)
                else arg
                for arg in args
            ]
            return super().__call__(*torch_filtered)

    numba.cuda.dispatcher._LaunchConfiguration = _BF16TorchWrappedLaunchConfiguration



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import os
from collections import defaultdict
import logging

from ast_canopy import parse_declarations_from_source
from numbast import (
    bind_cxx_enum,
    bind_cxx_structs,
    bind_cxx_functions,
    MemoryShimWriter,
)

from numba import types, config, cuda
from numba.core.datamodel.models import StructModel

from curand_device.curand_states import make_curand_states

logger = logging.getLogger("curand")

config.CUDA_USE_NVIDIA_BINDING = True

CUDA_INCLUDE_PATH = config.CUDA_INCLUDE_PATH
COMPUTE_CAPABILITY = cuda.get_current_device().compute_capability

curand_h = os.path.join(CUDA_INCLUDE_PATH, "curand.h")
curand_kernel_h = os.path.join(CUDA_INCLUDE_PATH, "curand_kernel.h")
curand_headers = [
    "curand_discrete2.h",
    "curand_discrete.h",
    "curand_globals.h",
    "curand.h",
    "curand_kernel.h",
    "curand_lognormal.h",
    "curand_mrg32k3a.h",
    "curand_mtgp32dc_p_11213.h",
    "curand_mtgp32.h",
    "curand_mtgp32_host.h",
    "curand_mtgp32_kernel.h",
    "curand_normal.h",
    "curand_normal_static.h",
    "curand_philox4x32_x.h",
    "curand_poisson.h",
    "curand_precalc.h",
    "curand_uniform.h",
]

curand_files = [os.path.join(CUDA_INCLUDE_PATH, f) for f in curand_headers]

curand_files += [os.path.normpath(p) for p in curand_files]

# Mysteriously, clang AST sometimes fails to stick to the one cuda include path
# for source range and instead uses the unwrapped path behind the symlink. E.g.
# /usr/local/cuda-12.3/include/curand_uniform.h instead of
# /usr/local/cuda/include/curand_uniform.h
with open(os.path.join(CUDA_INCLUDE_PATH, "cuda.h"), "r") as f:
    cuda_header_version_macro = [
        line for line in f if line.startswith("#define CUDA_VERSION")
    ]
    try:
        cuda_header_version = int(
            cuda_header_version_macro[0].split()[-1]
        )  # e.g. 12030
        major = cuda_header_version // 1000  # e.g. 12
        minor = (cuda_header_version % 1000) // 10  # e.g. 3

        curand_files += [
            os.path.join(f"/usr/local/cuda-{major}.{minor}/include/", f)
            for f in curand_headers
        ]
    except Exception as e:
        logger.warn(
            f"Failed to parse CUDA version from cuda.h, some functionality in curand may not be properly bound. {e}"
        )


curand_files = [h for h in curand_files if os.path.exists(h)]


structs, functions, _, _, typedefs, enums = parse_declarations_from_source(
    curand_kernel_h,
    curand_files,
    f"sm_{COMPUTE_CAPABILITY[0]}{COMPUTE_CAPABILITY[1]}",
    cudatoolkit_include_dir=CUDA_INCLUDE_PATH,
)


TYPE_SPECIALIZATION = {
    "curandDistributionShift_st": types.Type,
    "curandHistogramM2_st": types.Type,
    "curandDistributionM2Shift_st": types.Type,
    "curandDiscreteDistribution_st": types.Type,
    "curandStatePhilox4_32_10": types.Type,
    "normal_args_st": types.Type,
    "normal_args_double_st": types.Type,
    "curandStateTest": types.Type,
    "curandStateXORWOW": types.Type,
    "curandStateMRG32k3a": types.Type,
    "curandStateSobol32": types.Type,
    "curandStateScrambledSobol32": types.Type,
    "curandStateSobol64": types.Type,
    "curandStateScrambledSobol64": types.Type,
    "curandStateMtgp32": types.Type,
    "mtgp32_params_fast": types.Type,
    "mtgp32_kernel_params": types.Type,
}
DATA_MODEL_SPECIALIZATION = {
    "curandDistributionShift_st": StructModel,
    "curandHistogramM2_st": StructModel,
    "curandDistributionM2Shift_st": StructModel,
    "curandDiscreteDistribution_st": StructModel,
    "curandStatePhilox4_32_10": StructModel,
    "normal_args_st": StructModel,
    "normal_args_double_st": StructModel,
    "curandStateTest": StructModel,
    "curandStateXORWOW": StructModel,
    "curandStateMRG32k3a": StructModel,
    "curandStateSobol32": StructModel,
    "curandStateScrambledSobol32": StructModel,
    "curandStateSobol64": StructModel,
    "curandStateScrambledSobol64": StructModel,
    "curandStateMtgp32": StructModel,
    "mtgp32_params_fast": StructModel,
    "mtgp32_kernel_params": StructModel,
}

functions_to_ignore = {
    "_skipahead_scratch",
    "_skipahead_sequence_scratch",
    "__curand_generate_skipahead_matrix_xor",
    "_skipahead_inplace",
    "_skipahead_sequence_inplace",
}

numba_struct_types = []
numba_functions = []
numba_enums = []

aliases = defaultdict(list)
for typedef in typedefs:
    aliases[typedef.underlying_name].append(typedef.name)

shim_writer = MemoryShimWriter(
    f'#include "{curand_h}"\n' + f'#include "{curand_kernel_h}"\n'
)

# Enum type creation needs to happen before function and struct binding.
for e in enums:
    E = bind_cxx_enum(e)
    numba_enums.append(E)


numba_struct_types += bind_cxx_structs(
    shim_writer, structs, TYPE_SPECIALIZATION, DATA_MODEL_SPECIALIZATION, aliases
)

numba_functions += bind_cxx_functions(
    shim_writer, functions, exclude=functions_to_ignore
)


curandStates = [
    "curandStateTest",
    "curandStateXORWOW",
    "curandStateMRG32k3a",
    "curandStateSobol32",
    "curandStateScrambledSobol32",
    "curandStateSobol64",
    "curandStateScrambledSobol64",
    "curandStatePhilox4_32_10",
    # "curandStateMtgp32", # Require additional type parsing of mtgp32_params_fast and mtgp32_kernel_params
]

numpy_curand_states = []
states_arg_handlers = []

for state in curandStates:
    decl = next(s for s in structs if s.name == state)
    states_obj, arg_handler = make_curand_states(decl)
    numpy_curand_states.append(states_obj)
    states_arg_handlers.append(arg_handler)


# Export
globals().update({s.__name__: s for s in numba_struct_types})
globals().update({f.__name__: f for f in numba_functions})
globals().update({e.__name__: e for e in numba_enums})
globals().update({s.__name__: s for s in numpy_curand_states})
globals().update({"states_arg_handlers": states_arg_handlers})
for underlying_name, names in aliases.items():
    for name in names:
        if name not in globals() and underlying_name in globals():
            globals()[name] = globals()[underlying_name]
globals().update({"get_shims": shim_writer.links})

__all__ = list(  # noqa: F822
    set(s.__name__ for s in numba_struct_types)
    | set(f.__name__ for f in numba_functions)
    | set(typedef.name for typedef in typedefs)
    | set(e.__name__ for e in numba_enums)
    | set(s.__name__ for s in numpy_curand_states)
    | {"states_arg_handlers"}
) + ["get_shims"]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from numbast_extensions.bf16.bf16_bindings import *  # noqa: F403
from numbast_extensions.bf16.torch_interop import patch_numba

patch_numba()
